import pickle
import sys

from PIL import Image
from colormath.color_conversions import convert_color
from colormath.color_diff import delta_e_cie1976
from colormath.color_objects import LabColor, sRGBColor


def pixelate(filename, granularity, ncolors, nbits, colordiff_fn, verbose):
	try:
		im = Image.open(filename)
	except Exception as e:
		print(f"An error has occurred: {e}")
		sys.exit()

	width, height = im.size[0], im.size[1]
	print(f"Input image dimensions -> {width=}, {height=}")

	block_width = find_block_dim(granularity, width)  # find the possible block dimensions
	block_height = find_block_dim(granularity, height)
	print(f"Block dimensions -> {block_width=}, {block_height=}")

	grid_width = width // block_width  # dimension of the grid
	grid_height = height // block_height
	print(f"Grid dimensions -> {grid_width=}, {grid_height=}")

	nblocks = grid_width * grid_height  # number of blocks
	print(f"Number of blocks -> {nblocks=}")

	print(f"Fetching image blocks")
	blocks = []
	for n in range(nblocks):  # get all the image blocks
		blocks += [get_block(im, n, block_width, block_height)]

	print("Building pixelated image ...")
	new_image = im.copy()
	for n in range(int(nblocks)):
		if verbose:
			print(f"{(n / nblocks * 100):.2f}%")

		# define the target box where to paste the new block
		i = (n % grid_width) * block_width  # i,j -> upper left point of the target image
		j = n // grid_width * block_height
		box = (i, j, i + block_width, j + block_height)

		# compute the average color of the block
		colordiff = globals()[colordiff_fn]
		avg = avg_color(blocks[n], nbits, colordiff)

		# paste it
		new_image.paste(avg, box)

	pixelated_im = new_image.convert("P", palette=Image.ADAPTIVE, colors=ncolors)
	return pixelated_im


# find the dimension(height or width) according to the desired granularity (a lower granularity small blocks)
def find_block_dim(granularity, dim):
	assert (granularity > 0)
	candidate = 0
	block_dim = 1
	counter = 0
	while counter != granularity:  # while we dont achieve the desired granularity
		candidate += 1
		while ((dim % candidate) != 0):
			candidate += 1
			if candidate > dim:
				counter = granularity - 1
				break

		if candidate <= dim:
			block_dim = candidate  # save the current feasible length

		counter += 1

	assert (dim % block_dim == 0 and block_dim <= dim)
	return block_dim


# get a block of the image
def get_block(im, n, block_width, block_height):
	width = im.size[0]

	grid_width = width / block_width  # dimension of the grid

	i = (n % grid_width) * block_width  # i,j -> upper left point of the target block
	j = (n / grid_width) * block_height

	box = (i, j, i + block_width, j + block_height)
	block_im = im.crop(box)
	return block_im


# returns the average color of a given image
def avg_color(im, nbits, colordiff):
	avg_r = avg_g = avg_b = 0.0
	pixels = im.getdata()
	size = len(pixels)
	for p in pixels:
		avg_r += p[0] / float(size)
		avg_g += p[1] / float(size)
		avg_b += p[2] / float(size)

	if nbits != 24:
		try:
			palette = pickle.load(open("palette/" + str(nbits) + "bit.palette", "rb"))
		except Exception:
			print(f"An error has occurred while loading the color palette")
			sys.exit()
		best_color = palette[0]
		best_diff = colordiff(best_color, (avg_r, avg_g, avg_b))
		for i in range(1, len(palette)):
			candidate_diff = colordiff(palette[i], (avg_r, avg_g, avg_b))
			if candidate_diff < best_diff:
				best_diff = candidate_diff
				best_color = palette[i]

		return best_color
	else:
		return int(avg_r), int(avg_g), int(avg_b)


# calculate color difference of two pixels in the RGB space
# the less the better
def colordiff_rgb(pixel1, pixel2):
	delta_red = pixel1[0] - pixel2[0]
	delta_green = pixel1[1] - pixel2[1]
	delta_blue = pixel1[2] - pixel2[2]

	fit = delta_red ** 2 + delta_green ** 2 + delta_blue ** 2
	return fit


# http://python-colormath.readthedocs.org/en/latest/index.html
# calculate color difference of two pixels in the L*ab space
# the less the better
# pros: better results, cons: very slow
def colordiff_lab(pixel1, pixel2):
	# convert rgb values to L*ab values
	rgb_pixel_source = sRGBColor(pixel1[0], pixel1[1], pixel1[2], True)
	lab_source = convert_color(rgb_pixel_source, LabColor)

	rgb_pixel_palette = sRGBColor(pixel2[0], pixel2[1], pixel2[2], True)
	lab_palette = convert_color(rgb_pixel_palette, LabColor)

	# calculate delta e
	delta_e = delta_e_cie1976(lab_source, lab_palette)
	return delta_e


import argparse
import sys


def build_parser():
	parser = argparse.ArgumentParser(description='Pixel art', epilog="A fun toy.")
	parser.add_argument("-f", "--filename", nargs=1, help="input filename", required=True)
	parser.add_argument("-p", "--nbits", nargs=1, help="number of bits of the palette, default=24",
						choices=[3, 8, 9, 24], type=int, default=[24])
	parser.add_argument("-n", "--ncolors", nargs=1, help="number of colors to use: 1-256, default=256", type=int,
						default=[256])
	parser.add_argument("-g", "--granularity", nargs=1,
						help="granularity to be used (>0):  a bigger value means bigger blocks, default=1", type=int,
						default=[1])
	parser.add_argument("-l", "--labdiff", help="use *lab model, default=rgb", action='store_true', default=False)
	parser.add_argument("-v", "--verbose", help="show progress", action='store_true', default=False)
	parser.add_argument("-s", "--save", help="save the output image", action='store_true', default=False)

	return parser


def validate_args(parser, args):
	filename = args['filename'][0]
	nbits = args['nbits'][0]
	ncolors = args['ncolors'][0]
	granularity = args['granularity'][0]
	colordiff_fn = "colordiff_rgb" if not args['labdiff'] else "colordiff_lab"
	verbose =args["verbose"]
	save = args["save"]

	valid = True
	error_msg = ""
	if filename.split(".")[-1] == "png":
		error_msg = "File format not supported. Try with a .jpg"
		valid = False
	elif ncolors < 1 or ncolors > 256:
		error_msg = "Number of colors should be an integer between 1 and 256"
		valid = False
	elif granularity < 1:
		error_msg = "Granularity should be number greater or equal to 1"
		valid = False
	if not valid:
		if error_msg:
			sys.stdout.write(error_msg)
		parser.print_help()
		sys.exit()

	validated_args = {
		"filename": filename,
		"nbits": nbits,
		"ncolors": ncolors,
		"granularity": granularity,
		"colordiff_fn": colordiff_fn,
		"verbose": verbose,
		"save": save
	}
	return validated_args


def parse_args(parser):
	args = vars(parser.parse_args())
	args = validate_args(parser, args)
	return args



__name__ = 'pixel-artist'
__version__ = "0.0.2"
__author__ = "Alexandre Pinto"
__author_email__ = "alexandpinto@gmail.com"
__description__ = 'A command line utility to build pixelated images.'
__url__ = 'https://github.com/alexpnt/pixel-artist'
__docs__ = 'https://github.com/alexpnt/pixel-artist'
__issues__ = 'https://github.com/alexpnt/pixel-artist/issues'
__license__ = 'MIT'
__copyright__ = 'Copyright 2022 Alexandre Pinto'
__all__ = ["__version__", "__author__", "__author_email__", "__url__", "__license__"]



"""Main entry point for the pixelation tool."""

import sys

from .constants import SUCCESS
from .core import pixelate
from .parser import build_parser, parse_args


def main() -> None:
	"""Parses the command line arguments and bootstraps the project."""
	arg_parser = build_parser()
	args = parse_args(arg_parser)

	pixelated_image = pixelate(args['filename'],
							   args['granularity'],
							   args['ncolors'],
							   args['nbits'],
							   args["colordiff_fn"],
							   args['verbose'])
	pixelated_image.show()
	if args['save']:
		print("Saving to " + args['filename'].split(".")[0] + "_pixelated.png ...")
		pixelated_image.save(args['filename'].split(".")[0] + "_pixelated.png")

	return SUCCESS


if __name__ == "__main__":
	sys.exit(main())



import sys
from importlib.util import spec_from_file_location, module_from_spec
from pathlib import Path

from setuptools import find_packages, setup


def _load_module_from_src(mod_name, mod_fname, pkg_dir=""):
    cwd = Path.cwd()
    pkg_path = Path.joinpath(cwd, pkg_dir)
    mod_spec = spec_from_file_location(mod_name, f"{str(pkg_path)}/{mod_fname}.py")
    mod = module_from_spec(mod_spec)
    sys.modules[mod_name] = mod
    mod_spec.loader.exec_module(mod)
    return mod


about = _load_module_from_src("about", "__about__", "pixel_artist")
install_requires = [
    "Pillow==9.0.1",
    "colormath==3.0.0"
]

test_requires = []

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="pixel-artist",
    version=about.__version__,
    author=about.__author__,
    author_email=about.__author_email__,
    description=about.__description__,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url=about.__url__,
    license=about.__license__,
    packages=find_packages(where=".", exclude=["tests", "venv*"]),
    entry_points={
        'console_scripts': [
            'pixel-artist = pixel_artist.__main__:main',
        ],
    },
    python_requires=">=3.6",
    include_package_data=True,
    install_requires=install_requires,
    test_requires=test_requires,
    keywords=["python", "command-line", "console", "cmd", "pixel-art", "pixel", "image-processing"],
    project_urls={
        "Source": about.__url__,
        "Documentation": about.__docs__,
        "Bug Tracker": about.__issues__,
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Multimedia :: Graphics",
        "Topic :: Terminals",
        "Topic :: Utilities"
    ],
)
